// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#ifndef __ES_QWEN2_CLIENT_H__
#define __ES_QWEN2_CLIENT_H__

#include <iostream>
#include <string>
#include <map>
#include <unordered_map>
#include <vector>
#include <list>
#include <memory>
#include <utility>
#include <cstdint>
#include "es_npu_interface.h"
#include "es_sys_memory.h"
#include "llm_tokenizer.hpp"

typedef uint16_t fp16_t;
typedef int32_t prob_data_t;

#define QWEN2_CONFIG_FILE "/opt/eswin/sample-code/npu_sample/qwen_sample/src/qwen2_7b/config.json"
#define MASK_BUFFER_SIZE (sizeof(int16_t))

typedef struct {
    int32_t min;
    int32_t max;
} CLIENT_INTERVAL_S;

typedef struct {
    std::string modelDir;
    std::vector<std::string> blockModelFiles;
    bool logits;
    std::string logitsModelFile;
    std::string vocabFile;
    std::string vocabJson;
    std::string mergesFile;
    std::string tiktokenFile;
    std::string embeddingFile;
    int32_t vocabSize;
    int32_t embeddingSize;
    int32_t numHead;
    int32_t embeddingDim;
    int32_t embeddingDim1;
    int32_t embeddingDim2;
    int32_t tokenNum;
    int32_t tokenN1;
    int32_t tokenN2;
    int32_t precision;
    std::string language;
    std::vector<std::string> eosList;
    int32_t responseMaxLen;
    int32_t responseLineLen;
    std::map<int32_t, std::string> prompts;
    int32_t randomCnt;
    int32_t repeatLen;
    int32_t repeatCnt;
    int32_t logLevel;
    int32_t dataDump;
    CLIENT_INTERVAL_S dataTokenInterval;
    int32_t opDump;
    int32_t opType;
    CLIENT_INTERVAL_S opTokenInterval;
    CLIENT_INTERVAL_S opLayerInterval;
    CLIENT_INTERVAL_S opListInterval;
    int32_t perfEnable;
    int32_t modelTest;
    float repetitionPenalty;
    int32_t topK;
} CLIENT_CONFIG_S;

typedef enum {
    IN_TOKEN_IDX = 0,
    IN_MASK_IDX,
    IN_V_CACHE_IDX,
    IN_K_CACHE_IDX,
    IN_MAX_IDX
} INPUT_TENSOR_E;

typedef enum {
    OUT_HIDDEN_IDX = 0,
    OUT_V_CACHE_IDX,
    OUT_K_CACHE_IDX,
    OUT_MAX_IDX
} OUTPUT_TENSOR_E;

typedef struct {
    NPU_TASK_S task;
    void* inputMapAddr[IN_MAX_IDX];
    void* outputMapAddr[OUT_MAX_IDX];
} MODEL_TASK_S;

class Qwen2Client
{
 public:
    explicit Qwen2Client(const std::string& configPath);
    ~Qwen2Client();

    std::string genResponse(const std::string& prompts);
    std::map<int32_t, std::string> getOptionalPrompts();

 private:
    // Tokenization and embedding function for Chinese
    std::vector<std::string> splitUtf8(const std::string& str);
    std::vector<std::string> splitEn(const std::string& input);
    std::vector<std::string> splitSentence(const std::string& input);

    // Utils
    bool checkDuplicateWords(const std::string& sentence);
    fp16_t fp16Add(fp16_t a, fp16_t b);
    void printFp16Vec(const std::vector<fp16_t>& vec, int32_t num);
    ES_S32 dumpData(const void* address, const size_t size, const std::string& file);
    void displayLoadModelProgress(int32_t current, int32_t total);

    // Model loading and initialization
    ES_S32 loadConfig(const std::string& configPath);
    ES_S32 loadEmbedding();
    ES_S32 loadModel();
    ES_S32 prepareTensorBuffer(NPU_TENSOR_S &tensor, ES_DEV_BUF_S &devBuf, void **pAddr, bool cache);
    ES_S32 prepareBlockTaskBuffers(int32_t modelIdx, uint32_t modelId);
    ES_S32 prepareLogitsTaskBuffers(uint32_t modelId);
    ES_S32 prepareKVCacheBuffers(uint32_t modelId, MODEL_TASK_S &modelTask);
    ES_U64 calcKCacheOutputAddrOffset(int32_t num);
    ES_U64 calcVCacheOutputAddrOffset(int32_t num);
    prob_data_t getTopK(int32_t wordPosInVocab, const int32_t& maskVal, const std::vector<int32_t>& outputIds);
    void runModel(MODEL_TASK_S& modelTask);
    int32_t argmax(int16_t *data, int32_t size);
    std::string generate(const std::vector<ChatMessage> &messages);

    // Model release functions
    ES_S32 releaseTaskBuffers();

    inline bool checkInterval(int32_t value, const CLIENT_INTERVAL_S &interval);

    CLIENT_CONFIG_S mConfig;
    std::vector<MODEL_TASK_S> mBlockTasks;
    MODEL_TASK_S mLogitsTask;
    std::unique_ptr<QwenTokenizer> mTokenizer;
    int32_t mWordEmbeddingSize;
    fp16_t *mWordEmbeddings;
};

#endif